"
 COPYRIGHT (c) 1989 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Button subclass:#Toggle
       instanceVariableNames:'showLamp lampColor lampWidth lampHeight'
       classVariableNames:''
       poolDictionaries:''
       category:'Views-Interactors'
!

Toggle comment:'

COPYRIGHT (c) 1989 by Claus Gittinger
              All Rights Reserved

$Header: /cvs/stx/stx/libwidg/Toggle.st,v 1.5 1994-01-08 17:30:02 claus Exp $
written spring/summer 89 by claus
'!

!Toggle class methodsFor:'documentation'!

documentation
"
this button changes state whenever pressed and stays pressed until pressed
again. All the main action is in Button, Toggle just redefines buttonpress/
release behavior.
The toggle may optionally display a little kind-of-lamp (or LED), which
is turned on when the toggle is pressed. (i.e. as in the Interviews toolkit).

instance variables:
    showLamp    <Boolean>       true if a lamp should be displayed
    lampColor   <Color>         color of the lamp
    lampWidth   <Integer>       width of the lamp in pixel
    lampHeight  <Integer>       height of the lamp in pixel
"
! !

!Toggle methodsFor:'initialization'!

initStyle
    super initStyle.

    showLamp := resources name:'SHOW_LAMP' default:false.
    showLamp ifTrue:[
        onLevel := offLevel.
        lampColor := resources name:'LAMP_COLOR'
                            default:(Color red:100 green:100 blue:0). "yellow"
        lampWidth := (device horizontalPixelPerMillimeter * 1.8) rounded.
        lampHeight := (device verticalPixelPerMillimeter * 3.5) rounded.

        "dont know, if I like this ..."
        "
        activeBgColor := bgColor
        "
    ].
! !

!Toggle methodsFor:'accessing'!

lampColor:aColor
    "change the color of the toggle-lamp"

    lampColor := aColor.
    (shown and:[showLamp and:[pressed]]) ifTrue:[
        self redraw
    ]
! !


!Toggle methodsFor:'private'!

computeLabelOrigin
    "compute the origin of the form/text.
     redefined to move label to the right if there is a lamp."

    super computeLabelOrigin.
    showLamp ifTrue:[
        labelOriginX := labelOriginX + hSpace + lampWidth + hSpace.
    ]
!

computeLabelSize
    "compute the extent needed to hold the label plus the lamp"

    super computeLabelSize.
    showLamp ifTrue:[
        labelWidth := labelWidth + hSpace + lampWidth + hSpace.
        labelHeight := labelHeight max: lampHeight
    ]
! !

!Toggle methodsFor:'changing state'!

toggleNoAction
    "toggle, but do NOT perform any action - can be used to change a toggle
     under program control (i.e. turn one toggle off from another one)"

    pressed := pressed not.
    pressed ifTrue:[
        self level:onLevel.
    ] ifFalse:[
        self level:offLevel.
    ].
    shown ifTrue:[
        self redraw
    ]
!

toggle
    "toggle and perform the action"

    |action|

    enabled ifTrue:[
        self toggleNoAction.
        pressed ifTrue:[
            action := pressActionBlock
        ] ifFalse:[
            action := releaseActionBlock
        ].
        action notNil ifTrue:[action value]
    ]
! !

!Toggle methodsFor:'events'!

buttonPress:button x:x y:y
    button == 1 ifFalse:[
        ^ super buttonPress:button x:x y:y
    ].
    self toggle
!

buttonRelease:button x:x y:y
    button == 1 ifFalse:[
        ^ super buttonRelease:button x:x y:y
    ].
    "ignore"
! !

!Toggle methodsFor:'redrawing'!

drawWith:fg and:bg
    "redraw myself with fg/bg. Use super to draw the label, 
     drawing of the lamp here."

    |x y|

    super drawWith:fg and:bg.   "this draws the text"

    showLamp ifTrue:[
        x := hSpace + margin.
        y := (height - lampHeight) // 2.
        self drawEdgesForX:x y:y width:lampWidth height:lampHeight level:-1.
        pressed ifTrue:[
            self paint:lampColor.
        ] ifFalse:[
            self paint:bgColor.
        ].
        self fillRectangleX:x+2 y:y+2 width:lampWidth - 4 height:lampHeight - 4
    ]
! !
